/**
 * @vue-async/module-loader@1.0.6
 */

(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports, require('@vue-async/utils')) :
  typeof define === 'function' && define.amd ? define(['exports', '@vue-async/utils'], factory) :
  (global = global || self, factory(global.VueAsyncModuleLoader = {}, global.VueAsyncUtils));
}(this, (function (exports, utils) { 'use strict';

  const isProduction = "development" === 'production';
  const namespaces = 'dynamicComponent';
  // eslint-disable-next-line @typescript-eslint/no-unused-vars
  function dynamicComponent (Vue, store) {
      return {
          /**
           * add component to store
           * @param component vue component
           * @param position position
           */
          add(component, position) {
              if (!(typeof component === 'object' || typeof component === 'function')) {
                  return utils.error(isProduction, 'dynamicComponent function "add" only accept object or function component as param.');
              }
              store.commit(`${namespaces}/add`, {
                  component,
                  position,
              });
          },
          /**
           * remove component from store
           * @param name component name
           * @param position position
           */
          remove(name, position) {
              if (typeof name !== 'string') {
                  return utils.error(isProduction, 'dynamicComponent function "remove" only accept string name of component as param.');
              }
              store.commit(`${namespaces}/remove`, {
                  name,
                  position,
              });
          },
      };
  }

  const isProduction$1 = "development" === 'production';
  var dynamicComponentState = {
      namespaced: true,
      state: {
          GLOBAL: {},
      },
      mutations: {
          // 注册插槽组件
          add(state, { component, position = 'GLOBAL' }) {
              const _component = component.component || component;
              // 必须使用外部 Vue 对象，本地 import 对象不一致
              const componentName = component.name || utils.getComponentName(_component);
              if (!componentName) {
                  return utils.error(isProduction$1, 'component must be named for dynamic component.');
              }
              if (!state[position]) {
                  state[position] = {};
              }
              if (state[position][componentName]) {
                  return utils.error(isProduction$1, `async component "${componentName}" has existed in the ${position}, ignored it.`);
              }
              state[position][componentName] = component;
          },
          // 移除注册插槽组件
          remove(state, { name, position = 'GLOBAL' }) {
              if (!state[position]) {
                  return utils.error(isProduction$1, `position "${position}" does not  exist.`);
              }
              if (!state[position][name]) {
                  return utils.error(isProduction$1, `async component "${name}" does not  exist in the ${position}.`);
              }
              delete state[position][name];
          },
      },
  };

  const isProduction$2 = "development" === 'production';
  // eslint-disable-next-line @typescript-eslint/no-unused-vars
  function createEventBus (Vue) {
      let events = {};
      return {
          /**
           * 发射事件
           * @param {string} eventName 事件名称
           * @param {any} payload 事件载荷
           */
          emit(eventName, payload) {
              if (!events[eventName]) {
                  return utils.warn(isProduction$2, `event "${eventName}" is triggered, but no listeners，nothing will be happening.`);
              }
              for (const func of events[eventName].values()) {
                  func(payload);
              }
          },
          /**
           * 事件监听
           * @param eventName 事件名称
           * @param handler 处理事件的方法
           */
          on(eventName, handler) {
              if (typeof handler === 'function') {
                  if (!events[eventName])
                      events[eventName] = new Set();
                  events[eventName].add(handler);
              }
              else {
                  utils.error(isProduction$2, `handler must be a function, but recived ${typeof handler}`);
              }
          },
          /**
           * 取消事件监听
           * @param eventName 事件名称
           * @param handler 处理事件的方法
           */
          off(eventName, handler) {
              if (!handler) {
                  return utils.warn(!isProduction$2, 'handler must be a valid event function');
              }
              else if (!events.hasOwnProperty(eventName)) {
                  return utils.error(isProduction$2, `no event name of "${eventName}" is on listening`);
              }
              events[eventName].delete(handler);
          },
          /**
           * 清理事件总线
           */
          clear() {
              events = {};
          },
          /**
           * 获取当前事件总线详情
           */
          getEvents() {
              return events;
          },
      };
  }

  const isProduction$3 = "development" === 'production';
  var createModuleLoader = (Vue, status) => {
      return function loader(modules, { __isArray__ = false } = {}) {
          status.current = false;
          if (Array.isArray(modules) && modules.length) {
              return Promise.all(modules.map((module) => loader.call(this, module, { __isArray__: true }))).then(() => {
                  status.current = true;
              });
          }
          else if (utils.isPlainObject(modules)) {
              // resolve in the server render
              if (Vue.prototype.$isServer) {
                  if (!__isArray__) {
                      status.current = true;
                  }
                  return Promise.resolve();
              }
              /** 通过模块清单加载模块 */
              const promiseAll = [];
              for (const moduleName in modules) {
                  // todo: Duplicated module name
                  if (!window[moduleName]) {
                      promiseAll.push(new Promise((resolve) => {
                          const moduleConfig = modules[moduleName];
                          // load css
                          if (utils.isPlainObject(moduleConfig) && moduleConfig.css) {
                              (Array.isArray(moduleConfig.css) ? moduleConfig.css : [moduleConfig.css]).map((href) => {
                                  const link = document.createElement('link');
                                  link.rel = 'stylesheet';
                                  link.type = 'text/css';
                                  link.href = href;
                                  link.dataset.styleFor = moduleName;
                                  document.getElementsByTagName('head')[0].appendChild(link);
                              });
                          }
                          // load module
                          const entry = typeof moduleConfig === 'string' ? moduleConfig : moduleConfig.entry;
                          const script = document.createElement('script');
                          script.src = entry;
                          script.onload = () => {
                              const module = window[moduleName];
                              if (module) {
                                  utils.isFunction(module)
                                      ? module.call(this, Vue)
                                      : typeof module.default === 'function' // requirejs
                                          ? module.default.call(this, Vue)
                                          : utils.warn(isProduction$3, `module "${moduleName}" entry must be a function，ignored it.`);
                                  {
                                      utils.print(`module:${moduleName}`, 'loaded');
                                  }
                              }
                              else {
                                  utils.warn(isProduction$3, `module "${moduleName}" loaded failed，ignored it.`);
                              }
                              resolve();
                          };
                          script.onerror = () => {
                              utils.warn(isProduction$3, `module "${moduleName}" had a problem to create, ignored it.`);
                              resolve();
                          };
                          document.body.appendChild(script);
                      }));
                  }
              }
              return Promise.all(promiseAll).then(() => {
                  if (!__isArray__) {
                      status.current = true;
                  }
              });
          }
          else if (typeof modules === 'function') {
              /** 通过模块函数加载模块 */
              const result = modules.call(this, Vue);
              if (result && result instanceof Promise) {
                  return result.then(() => {
                      if (!__isArray__) {
                          status.current = true;
                      }
                  });
              }
              else {
                  if (!__isArray__) {
                      status.current = true;
                  }
                  return Promise.resolve();
              }
          }
          else {
              utils.error(isProduction$3, '模块加载方法只接受模块列表对象或者模块函数对象作为参数。');
              if (!__isArray__) {
                  status.current = true;
              }
              return Promise.resolve();
          }
      };
  };

  var createComponentLoader = () => {
      return function loader(componentName, url) {
          return new Promise((resolve, reject) => {
              let component = window[componentName];
              // exists
              if (component) {
                  resolve(component);
              }
              else {
                  const script = document.createElement('script');
                  script.src = url;
                  script.async = true;
                  script.onload = () => {
                      component = window[componentName];
                      // check again after js loaded
                      if (component) {
                          {
                              utils.print(`component:${componentName}`, 'loaded');
                          }
                          resolve(component);
                      }
                      else {
                          reject(new Error(`component "${componentName}" loaded failed. `));
                      }
                  };
                  script.onerror = () => {
                      reject(new Error(`component "${componentName}" had a problem to load. ${url}`));
                  };
                  document.body.appendChild(script);
              }
          });
      };
  };

  function install(Vue) {
      if (install.installed)
          return;
      install.installed = true;
      // 用于监听模块加载的情况，处理路由404问题
      const vm = new Vue({
          data() {
              return {
                  status: {
                      current: false,
                  },
              };
          },
      });
      Object.defineProperties(Vue.prototype, {
          $eventBus: {
              value: createEventBus(),
              writable: false,
          },
          $moduleLoader: {
              value: createModuleLoader(Vue, vm.status),
              writable: false,
          },
          $componentLoader: {
              value: createComponentLoader(),
              writable: false,
          },
      });
      // router
      const routerInject = (router) => {
          // 解决动态路由404问题
          const resolveRoute = (to, from, next) => {
              const fullPath = to.redirectedFrom || to.fullPath;
              const { resolved, location } = router.resolve(fullPath);
              // 在加载完组件后resolve的地址与原来要跳转的地址不一致时跳转
              // 以免造成404死循环
              if (resolved.name !== to.name) {
                  next(location);
              }
              else {
                  next();
              }
          };
          router.beforeEach((to, from, next) => {
              if (!to.name || to.name === '404' || to.name === 'page-not-found' || to.path === '*') {
                  // 模块已经被加载完成, 但由于在其之前添加 beforeEach 阻止时间过长，vm.$watch还没开始监听
                  if (vm.status.current) {
                      resolveRoute(to, from, next);
                  }
                  vm.$watch(() => vm.status.current, (newVal, oldVal) => {
                      // false => true
                      if (newVal && !oldVal) {
                          resolveRoute(to, from, next);
                      }
                  });
              }
              else {
                  next();
              }
          });
      };
      // store
      const storeInject = (store) => {
          store.registerModule(namespaces, dynamicComponentState);
          // define $dynamicComponent
          Object.defineProperty(Vue.prototype, '$dynamicComponent', {
              value: dynamicComponent(Vue, store),
              writable: false,
          });
      };
      const _init = Vue.prototype._init;
      Vue.prototype._init = function (options = {}) {
          // 从 Vue root option 中获取 router | store 实例
          // router beforeEach 需要在 beforeCreate 之前添加，才能在页面强制刷新时第一次生效
          if (options.router) {
              routerInject(options.router);
          }
          if (options.store) {
              storeInject(options.store);
          }
          _init.call(this, options);
      };
      // Used to avoid multiple mixins being setup
      // when in dev mode and hot module reload
      // https://github.com/vuejs/vue/issues/5089#issuecomment-284260111
      if (Vue.$__module_loader_installed__)
          return;
      // eslint-disable-next-line @typescript-eslint/camelcase
      Vue.$__module_loader_installed__ = true;
      if (window && !window.Vue) {
          window.Vue = Vue;
      }
      Vue.mixin({
          beforeCreate() {
              const options = this.$options;
              if (options.moduleLoader) {
                  options.moduleLoader.init(this, options.ssrContext);
                  this.$moduleLoadManager = Vue.observable(options.moduleLoader.framework);
              }
              else {
                  this.$moduleLoadManager = (options.parent && options.parent.$moduleLoadManager) || this;
              }
          },
      });
  }

  class ModuleLoader {
      constructor(options) {
          this._framework = {
              layouts: {},
          };
          this._options = options;
      }
      get framework() {
          return this._framework;
      }
      // eslint-disable-next-line @typescript-eslint/no-unused-vars
      init(root, ssrContext) {
          this._options &&
              Object.entries(this._options).map(([key, value]) => {
                  if (!utils.hasOwn(this._framework, key)) {
                      this._framework[key] = value;
                  }
                  else {
                      utils.warn("development" === 'production', `参数 ${key} 已在在`);
                  }
              });
          // 默认方法
          !utils.hasOwn(this._framework, 'addRoutes') && (this._framework.addRoutes = this._createAddRoutes(root));
          !utils.hasOwn(this._framework, 'addLayouts') && (this._framework.addLayouts = this._createAddLayouts(root));
      }
      _createAddRoutes(root) {
          return (routes) => {
              if (root.$router) {
                  root.$router.addRoutes(routes);
              }
              else {
                  utils.error("development" === 'production', 'vm.$router未定义');
              }
          };
      }
      _createAddLayouts(root) {
          return (key, layout) => {
              if (typeof key === 'string') {
                  root.$set(this._framework.layouts, key, layout);
              }
              else if (utils.isPlainObject(key)) {
                  // plain object
                  const layouts = key;
                  this._framework.layouts = Object.assign(this._framework.layouts, layouts);
              }
              else {
                  utils.error("development" === 'production', 'addLayouts中支持字符串(key)或 Object(key/value)格式');
              }
          };
      }
  }
  ModuleLoader.install = install;
  ModuleLoader.installed = false;
  ModuleLoader.version = '1.0.6';

  const install$1 = ModuleLoader.install;
  ModuleLoader.install = (Vue) => {
      install$1.call(ModuleLoader, Vue);
  };
  // Auto install if it is not done yet and `window` has `Vue`.
  // To allow users to avoid auto-installation in some cases,
  if (window && window.Vue) {
      window.Vue.use(ModuleLoader);
  }

  exports.default = ModuleLoader;

  Object.defineProperty(exports, '__esModule', { value: true });

})));
